from setuptools import setup

setup(
    name='mltest',
    version='0.1.3',
    packages=['mltest'],
    license="MIT",
    author="Chase Roberts",
    url="https://github.com/Thenerdstation/mltest",
    keywords=['testing', 'machine learning'],
    author_email="keeper6928@gmail.com"
)
